from adam_atan2_pytorch.adam_atan2 import AdamAtan2
from adam_atan2_pytorch.adopt_atan2 import AdoptAtan2

from adam_atan2_pytorch.muon_adam_atan2 import MuonAdamAtan2

Adam = AdamAtan2
Adopt = AdoptAtan2
